"""
Gimbal -- Properties and transformations of 3D molecular geometries.
"""
from gimbal.molecule import import_molecule
from gimbal.molecule import import_bundle
from gimbal._version import __version__


__author__ = 'Ryan J. MacDonell'
__license__ = 'MIT'
