"""
Tests for the symmetry module.
"""
import numpy as np
import gimbal.symmetry as symmetry


def test_principle_axes():
    pass


def test_assign_rotor():
    pass


def test_symmetrize():
    pass


def test_cart_jumble():
    pass
